/*
 * Copyright (C) 2001-2007 by egnite Software GmbH
 * Copyright (C) 2008-2009 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*!
 * \file httpserv.c
 * \brief HTTP interface.
 *
 * \verbatim
 * $Id: httpserv.c 47 2010-06-09 10:45:51Z haraldkipp $
 * \endverbatim
 */

#include "config.h" /* Must be first. */

#if USE_HTTPSERVER

#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>

#include <dev/board.h>
#include <dev/urom.h>
#include <dev/vscodec.h>
#include <dev/watchdog.h>

#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <sys/confos.h>
#include <sys/confnet.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/heap.h>
#include <sys/bankmem.h>
#include <sys/socket.h>
#include <sys/version.h>

#include <pro/ssi.h>
#include <pro/httpd.h>

#include <memdebug.h>

#include "logmsg.h"
#include "utils.h"
#include "xmlutils.h"
#include "favlist.h"
#include "proxy.h"
#include "webradio.h"
#include "shoutcast.h"
#include "httpserv.h"


/* Service thread counter. */
static int httpd_tc;

static int StartServiceThread(void);

static int GetEncodedHexChar(FILE *stream)
{
    uint_fast8_t i;
    uint_fast8_t c = 0;
    int ch;

    for (i = 0; i < 2; i++) {
        c <<= 4;
        ch = fgetc(stream);
        if (ch >= '0' && ch <= '9') {
            c |= ch - '0';
        }
        else if (ch >= 'A' && ch <= 'F') {
            c |= ch - ('A' - 10);
        }
        else if (ch >= 'a' && ch <= 'f') {
            c |= ch - ('a' - 10);
        } 
        else {
            LogMsg(LOG_SHOUTCAST, "Bad hex code %d\r\n", ch);
            return EOF;
        }
    }
    return (int)c;
}

int GetEncodedString(FILE * stream, char *str, size_t siz, int eos)
{
    int ch;
    size_t s;

    for (s = 1; s < siz; s++) {
        ch = fgetc(stream);
        if (ch == EOF || ch == eos) {
            if (str) {
                *str = '\0';
            }
            return (int)s - 1;
        }
        if (ch == '+') {
            ch = ' ';
        }
        else if (ch == '%') {
            ch = GetEncodedHexChar(stream);
            if (ch == EOF) {
                break;
            }
        }
        if (str) {
            *str++ = (char)ch;
        }
    }
    return -1;
}

void AddStationUris(int idx, CONST char *str)
{
    char *uri = malloc(256);
    CONST char *cp;
    char *up;

    if (uri) {
        cp = str;
        while (*cp) {
            while (*cp && *cp < ' ') {
                cp++;
            }
            up = uri;
            while (*cp) {
                if (*cp < ' ') {
                    break;
                }
                *up++ = *cp++;
            }
            *up = 0;
            if (strlen(uri)) {
                LogMsg(LOG_HTTPD, "Add '%s'\r\n", uri);
                FavListSet(idx, NULL, uri);
            }
        }
        free(uri);
    }
}

/*! \fn Service(void *arg)
 * \brief HTTP service thread.
 *
 * The endless loop in this thread waits for a client connect,
 * processes the HTTP request and disconnects. Nut/Net doesn't
 * support a server backlog. If one client has established a
 * connection, further connect attempts will be rejected.
 * Typically browsers open more than one connection in order
 * to load images concurrently. So we run this routine by
 * several threads.
 *
 */
THREAD(Service, arg)
{
    TCPSOCKET *sock;
    FILE *stream;

    LogMsg(LOG_HTTPD, "Started HTTP daemon\r\n");
    for (;;) {
        /* Create a socket. */
        if ((sock = NutTcpCreateSocket()) == 0) {
            LogMsg(LOG_WARN, "No sockets\r\n");
            NutSleep(1000);
            continue;
        }

        /* Set socket options. Silently ignore any error. */
#ifdef HTTPD_MAX_SEGSIZE
        {
            uint16_t mss = HTTPD_MAX_SEGSIZE;
            NutTcpSetSockOpt(sock, TCP_MAXSEG, &mss, sizeof(mss));
        }
#endif
#ifdef HTTPD_TCP_BUFSIZE
        {
            uint16_t tcpbufsiz = HTTPD_TCP_BUFSIZE;
            NutTcpSetSockOpt(sock, SO_RCVBUF, &tcpbufsiz, sizeof(tcpbufsiz));
        }
#endif

        stream = TcpStreamAccept(sock, HTTPD_TCP_PORT, "r+b");
        if (stream) {
            LogDbgMsg(LOG_HTTPD, "Connected HTTP client\r\n");
#if HTTPD_TCP_TIMEOUT
            {
                uint32_t tmo = HTTPD_TCP_TIMEOUT;
                NutTcpSetSockOpt(sock, SO_RCVTIMEO, &tmo, sizeof(tmo));
            }
#endif
#if USE_DYNAMIC_THREADS
            /* Start a new thread and let it run. */
            StartServiceThread();
            NutSleep(1);
#endif
            /* Let the Nut/OS library process the client's requests. */
            NutHttpProcessRequest(stream);
            LogDbgMsg(LOG_HTTPD, "Disconnecting HTTP client\r\n");
            fclose(stream);
        }
        NutTcpCloseSocket(sock);

#if USE_DYNAMIC_THREADS
        /* If enough threads are running, stop this one. */
        if (httpd_tc >= HTTPD_MIN_THREADS) {
            httpd_tc--;
            LogDbgMsg(LOG_HTTPD, "Stop HTTP daemon\r\n");
            NutThreadExit();
        }
#endif
    }
}

/*
 * Start a HTTP daemon thread.
 */
static int StartServiceThread(void)
{
#if USE_DYNAMIC_THREADS
    if (httpd_tc >= HTTPD_MAX_THREADS) {
        return 0;
    }
#endif

    if (NutThreadCreate("httpd", Service, NULL, HTTPD_SERVICE_STACK) == NULL) {
        return -1;
    }
    httpd_tc++;

    return 0;
}

/*!
 * \brief HTTP Server Start.
 */
int HttpServerStart(void)
{
    int i;

    /*
     * Register our device for the file system.
     */
    if (NutRegisterDevice(&RADIO_FSDEV_HTTP, 0, 0)) {
        LogMsg(LOG_ERROR | LOG_HTTPD, "No HTTP mount\r\n");
        return -1;
    }

#ifdef RADIO_HTTP_ROOT
    /* Register non-default root directory. */
    LogMsg(LOG_HTTPD, "Register root dir '" RADIO_HTTP_ROOT "'\r\n");
    if (NutRegisterHttpRoot(RADIO_HTTP_ROOT)) {
        LogMsg(LOG_ERROR | LOG_HTTPD, "No HTTP root\r\n");
        return -1;
    }
#endif
    /*
     * Register SSI, CGI and JavaScript handlers.
     */
    NutRegisterSsi();
    RegisterMajs();
    RegisterCgis();

    /*
     * Start four server threads.
     */
    LogDbgMsg(LOG_HTTPD, "Starting %d HTTP daemons\r\n", HTTPD_MIN_THREADS);
    for (i = 0; i < HTTPD_MIN_THREADS; i++) {
        if (StartServiceThread()) {
            LogMsg(LOG_ERROR | LOG_HTTPD, "Thread start failed\r\n");
            return -1;
        }
    }

    /* Let started threads take over. */
    NutSleep(1);

    return 0;
}

#endif /* USE_HTTPSERVER */
