/*
 * Copyright (C) 2008-2010 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * \file xmlutils.c
 * \brief XML utility routines.
 *
 * \verbatim
 * $Id$
 * \endverbatim
 */

#include "config.h" /* Must be first. */

#if USE_HTTPSERVER

#include <sys/heap.h>
#include <sys/timer.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <memdebug.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <pro/uxml.h>

#include "logmsg.h"
#include "proxy.h"
#include "utils.h"
#include "xmlutils.h"

#ifndef MAX_XMLTAG_SIZE
#define MAX_XMLTAG_SIZE     128
#endif

#ifndef MAX_XMLTKN_SIZE
#define MAX_XMLTKN_SIZE     32
#endif

/*
 * Unfortunately the UxmlReadTag routine is static. 
 * We have to add a local copy.
 */
static int XmlReadTag(FILE * stream, char *data, size_t size)
{
    int rc = -1;
    int ch;
    int qc = 0;
    int state = 1;
    char *dp = NULL;

    while (state) {
        ch = fgetc(stream);
        if (ch == EOF || ch == 0) {
            break;
        }
        switch (state) {
        case 1:
            /* Searching first bracket. */
            if (ch == '<') {
                /* Opening bracket found. Start collecting. */
                dp = data;
                state = 4;
            } else if (ch == '"' || ch == '\'') {
                /* Quote found. Skip quoted string. */
                qc = ch;
                state++;
            }
            break;
        case 2:
            /* Skipping quoted string. */
        case 5:
            /* Collecting quoted string. */
            if (ch == qc) {
                /* End quote found. */
                state--;
            }
            break;
        case 3:
            /* Compressing spaces. */
            if (isspace(ch)) {
                ch = 0;
                break;
            }
            state = 4;
            /* Fall through. */
        case 4:
            /* Collecting data. */
            if (ch == '>') {
                rc = 0;
                state = 0;
            } else if (ch == '"' || ch == '\'') {
                qc = ch;
                state++;
            } else if (isspace(ch)) {
                ch = ' ';
                state = 3;
            }
            break;
        }
        if (dp && ch) {
            if (size > 1) {
                size--;
                *dp++ = ch;
            } else {
                break;
            }
        }
    }
    if (dp) {
        *dp = 0;
    }
    return rc;
}

/*!
 * \brief Collect a string list of attributes from an XML stream.
 *
 * \param stream The stream to read from.
 * \param f_tag  The tag to search for.
 * \param f_att  The attribute to collect.
 *
 * \return Pointer to a newly allocated string list on success.
 *         Otherwise a NULL pointer is returned.
 */
char * XmlParseStreamStringList(FILE * stream, char *f_tag, char *f_att)
{
    char *tag;
    char *tkn;
    int tkn_len;
    char *tp;
    char *list = NULL;
    int list_size = 0;
    int list_len = 0;

    /* Allocate the tag buffer. */
    tag = malloc(MAX_XMLTAG_SIZE);
    if (tag == NULL) {
        return NULL;
    }
    /* Allocate the token buffer. */
    tkn = malloc(MAX_XMLTKN_SIZE);
    if (tkn == NULL) {
        free(tag);
        return NULL;
    }
    for (;;) {
        if (NutHeapAvailable() < 6144) {
            break;
        }
        /* Read the next tag. */
        if (XmlReadTag(stream, tag, MAX_XMLTAG_SIZE)) {
            /* No more tags or error. */
            break;
        }

        /* Parse the tag. */
        tp = UxmlParseTag(tag + 1, tkn, MAX_XMLTKN_SIZE);
        if (tp) {
            if (isalpha((unsigned char)*tkn) && strcasecmp(tkn, f_tag) == 0) {
                /* Parse the attributes. */
                for (;;) {
                    tp = UxmlParseTag(tp, tkn, MAX_XMLTKN_SIZE);
                    if (tp == NULL || *tkn == '>') {
                        /* End of this tag or error. */
                        break;
                    }
                    if (isalpha((unsigned char)*tkn) && strcasecmp(tkn, f_att) == 0) {
                        char *name = strdup(tkn);

                        if (name) {
                            tp = UxmlParseTag(tp, tkn, MAX_XMLTKN_SIZE);
                            if (tp == NULL || *tkn != '=') {
                                free(name);
                                break;
                            }
                            tp = UxmlParseTag(tp, tkn, MAX_XMLTKN_SIZE);
                            if (tp == NULL || *tkn == '>') {
                                free(name);
                                break;
                            }
                            tkn_len = strlen(tkn) + 1;
                            /* Allocate the string list. */
                            if (list_len + tkn_len > list_size) {
                                list_size += 128;
                                list = realloc(list, list_size);
                                if (list == NULL) {
                                    break;
                                }
                            }
                            strcpy(list + list_len, tkn);
                            list_len += tkn_len;
                            list[list_len] = '\0';
                            free(name);
                        }
                    }
                }
            } else if (*tkn == '/') {
                /*
                 * End of the active node.
                 */
                tp = UxmlParseTag(tp, tkn, MAX_XMLTKN_SIZE);
            }
        }
    }
    /* Clean up. */
    free(tag);
    free(tkn);

    return list;
}

#define MAX_UXMLTKN_SIZE    128
#define MAX_UXMLTAG_SIZE    512

extern FILE *OpenTcpStream(TCPSOCKET *sock, CONST char *host, uint16_t port);
extern int SendHttpRequest(FILE *stream, HTTP_SCHEME *schm);

XMLSTREAM *XmlStreamOpen(char *uri)
{
    printf("Connecting '%s'\n", uri);
    /* Split URI into scheme structure. */
    HTTP_SCHEME *schm = HttpSchemeParse(uri);

    if (schm) {
        /* Create a TCP socket to connect the SHOUTcast server. */
        TCPSOCKET *sock = NutTcpCreateSocket();

        if (sock) {
            /* Connect the SHOUTcast server and associate a stream. */
            FILE *stream = OpenTcpStream(sock, schm->schm_host, schm->schm_portnum);

            if (stream) {
                XMLSTREAM *xs;

                /* Use the stream to transmit our request. */
                SendHttpRequest(stream, schm);

                xs = malloc(sizeof(XMLSTREAM));
                if (xs) {
                    HttpSchemeRelease(schm);
                    xs->sock = sock;
                    xs->stream = stream;
                    /* Allocate the tag buffer. */
                    xs->tag = malloc(MAX_UXMLTAG_SIZE);
                    /* Allocate the token buffer. */
                    xs->tkn = malloc(MAX_UXMLTKN_SIZE);
                    if (xs->tag && xs->tkn) {
                        return xs;
                    }
                    free(xs->tag);
                    free(xs->tkn);
                    free(xs);
                }
                fclose(stream);
            } else {
                printf("Failed to open stream\n");
            }
            NutTcpCloseSocket(sock);
        }
        /* Release the scheme structure. */
        HttpSchemeRelease(schm);
    }
    return NULL;
}

#include <ctype.h>
extern int XmlReadTag(FILE * stream, char *data, size_t size);

char **XmlStreamGet(XMLSTREAM *xs, char *f_tag, char *f_attr[], int n_attr)
{
    char *tp;
    int idx;
    char **attr = NULL;

    while (attr == NULL) {
        /* Read the next tag. */
        if (XmlReadTag(xs->stream, xs->tag, MAX_UXMLTAG_SIZE)) {
            /* No more tags or error. */
            return attr;
        }

        /* Parse the tag. */
        tp = UxmlParseTag(xs->tag + 1, xs->tkn, MAX_UXMLTKN_SIZE);
        if (tp) {
            if (strcasecmp(xs->tkn, f_tag) == 0) {
                /* Parse the attributes. */
                for (;;) {
                    tp = UxmlParseTag(tp, xs->tkn, MAX_UXMLTKN_SIZE);
                    if (tp == NULL || *xs->tkn == '>') {
                        /* End of this tag or error. */
                        break;
                    }
                    for (idx = 0; idx < n_attr; idx++) {
                        if (strcasecmp(xs->tkn, f_attr[idx]) == 0) {
                            break;
                        }
                    }
                    if (idx < n_attr) {
                        tp = UxmlParseTag(tp, xs->tkn, MAX_UXMLTKN_SIZE);
                        if (tp == NULL || *xs->tkn != '=') {
                            break;
                        }
                        tp = UxmlParseTag(tp, xs->tkn, MAX_UXMLTKN_SIZE);
                        if (tp == NULL || *xs->tkn == '>') {
                            break;
                        }
                        if (attr == NULL) {
                            attr = malloc(n_attr * sizeof(char *));
                            if (attr == NULL) {
                                return NULL;
                            }
                        }
                        attr[idx] = strdup(xs->tkn);
                    }
                }
            }
            else if (*xs->tkn == '/') {
                /* End of tag. */
                tp = UxmlParseTag(tp, xs->tkn, MAX_UXMLTKN_SIZE);
            }
        }
    }
    return attr;
}

void XmlStreamClose(XMLSTREAM *xs)
{
    /* Close the connection. */
    fclose(xs->stream);
    NutTcpCloseSocket(xs->sock);
    free(xs->tag);
    free(xs->tkn);
    free(xs);
}

#endif /* USE_HTTPSERVER */
